@@ -1,4 +1,8 @@
-from abc import ABC
+from __future__ import annotations
+
+import pickle
+import struct
+from dataclasses import dataclass, fields, is_dataclass
 from pathlib import Path
 from typing import (
     Callable,
@@ -7,27 +11,25 @@ from typing import (
     Literal,
     Optional,
     Sequence,
-    Set,
+    Tuple,
     Type,
     TypeVar,
     Union,
     get_origin,
     get_type_hints,
-    Tuple,
 )
-from dataclasses import dataclass, fields, is_dataclass
-from nptyping.ndarray import NDArrayMeta  # type: ignore
-from nptyping.shape_expression import get_dimensions  # type: ignore
+from copy import deepcopy
 import numpy as np
 import orjson
-import pyarrow.parquet as pq
 import pyarrow as pa
-import pickle
-import struct
+import pyarrow.parquet as pq
+from nptyping.ndarray import NDArrayMeta  # type: ignore
+from nptyping.shape_expression import get_dimensions  # type: ignore
 from ream.helper import has_dict_with_nonstr_keys
 from serde.helper import get_compression, get_open_fn
 from tqdm import tqdm
 
+
 T = TypeVar("T")
 
 
@@ -311,88 +313,6 @@ class NumpyDataModel:
 
         return cls(**kwargs)
 
-    def is_range_index_valid(self):
-        """Check if the index of a numpy data model is valid: the index is contiguous, no overlapping
-        and covers the entire range of the numpy arrays.
-
-        This function assumes that the index is a dictionary (or instance of Index[dict]), and each value in the index
-        is either (start, end) if the index is a single-level index or (start, end, nested_index) if the index is a
-        multi-level index.
-        """
-        metadata: NumpyDataModelMetadata = self._metadata  # type: ignore
-        size = len(self)
-        for name in metadata.index_props:
-            index = getattr(self, name)
-            if isinstance(index, Index):
-                index = index.index
-            if not isinstance(index, dict):
-                raise TypeError(
-                    f"Index {name} is not a dictionary. This function only supports dictionary index"
-                )
-
-            # figure out the index level
-            index_level = self._get_range_index_level(index)
-            if index_level == 0 and size != 0:
-                raise ValueError("Index is empty but the numpy array is not empty")
-
-            indices = [index]
-            for level in tqdm(
-                range(index_level), desc="check index at different levels"
-            ):
-                cic = ContiguousIndexChecker()
-                next_indices = []
-                for idx in tqdm(indices):
-                    if len(idx) == 0:
-                        continue
-                    for elem in idx.values():
-                        cic.next(elem[0], elem[1])
-                        if len(elem) == 3:
-                            next_indices.append(elem[2])
-                if not cic.start == size:
-                    raise ValueError(
-                        f"Index is not covered all elements in the array at level {level}. Expected {size} but got {cic.start}"
-                    )
-                indices = next_indices
-
-    def _get_range_index_level(self, index: dict):
-        """Get the level of a range index. This function assumes that the index is a dictionary, and each value in the index
-        is either (start, end) if the index is a single-level index or (start, end, nested_index) if the index is a
-        multi-level index."""
-        if len(index) == 0:
-            return 0
-
-        level = []
-        for key, value in index.items():
-            if not isinstance(value, (list, tuple)):
-                raise TypeError(f"Index element {value} is not a sequence")
-            if len(value) == 2:
-                if not isinstance(value[0], int) or not isinstance(value[1], int):
-                    raise TypeError(
-                        f"Index element {value} does not have the form: (start, end)"
-                    )
-                level.append(1)
-            elif len(value) == 3:
-                if (
-                    not isinstance(value[0], int)
-                    or not isinstance(value[1], int)
-                    and not isinstance(value[2], dict)
-                ):
-                    raise TypeError(
-                        f"Index element {value} does not have the form: (start, end, nested_index)"
-                    )
-                level.append(self._get_range_index_level(value[2]) + 1)
-            else:
-                raise TypeError(
-                    f"Index element {value} does not have the form: (start, end) or (start, end, nested_index)"
-                )
-
-        level = set(level)
-        if len(level) > 1:
-            raise ValueError(
-                "The index is not a balanced multi-level index as some elements have different levels"
-            )
-        return level.pop()
-
 
 @dataclass
 class NumpyDataModelContainer:
@@ -432,3 +352,268 @@ class ContiguousIndexChecker:
             )
         self.start = end
         return self
+
+
+class NumpyDataModelHelper:
+    """Generally contains helpers for a numpy data model that has a single index, which is a dictionary.
+    The dictionary can have multiple levels, each level store a tuple of (start, end, nested_index),
+    and the dictionaries at the lowest level store just (start, end).
+
+    However, some helper methods can be used for other cases as well.
+    """
+
+    @classmethod
+    def stack(cls, npmodels: list[NumpyDataModel], keys: Sequence[str | int]):
+        """This method concatenates a list of numpy data models of the same type into a single numpy data model.
+        The new index is a new dictionary that each key provided maps to the index of each model in the provided order.
+        """
+        if len(npmodels) == 0:
+            raise ValueError("The provided list of numpy data models is empty")
+
+        metadata: NumpyDataModelMetadata = npmodels[0]._metadata  # type: ignore
+        assert len(set(npmodel.__class__ for npmodel in npmodels)) == 1
+        assert len(metadata.index_props) == 1
+
+        arrays = {prop: [] for prop in metadata.array_props}
+        index = {}
+        counter = 0
+        for npmodel, key in zip(npmodels, keys):
+            for prop, values in arrays.items():
+                arr = getattr(npmodel, prop)
+                values.append(arr)
+                size = len(arr)
+            size = len(next(iter(arrays.values()))[-1])
+            npmodel_index = deepcopy(getattr(npmodel, metadata.index_props[0]))
+            cls.offset_index(npmodel_index, counter)
+            index[key] = (counter, counter + size, npmodel_index)
+            counter += size
+
+        for prop, arr in arrays.items():
+            arrays[prop] = np.concatenate(arr)
+        arrays[metadata.index_props[0]] = index
+        return npmodels[0].__class__(**arrays)
+
+    @classmethod
+    def to_df(cls, npmodel: NumpyDataModel):
+        """This method converts a numpy data model to a pandas dataframe. Assuming that the index is a dictionary and it only has one index"""
+        import pandas as pd
+
+        metadata: NumpyDataModelMetadata = npmodel._metadata  # type: ignore
+        assert len(metadata.index_props) <= 1
+        if len(metadata.index_props) == 1:
+            index = getattr(npmodel, metadata.index_props[0])
+            assert isinstance(index, dict)
+            pdindex = pd.MultiIndex.from_tuples(cls.index_to_array(index))
+            return pd.DataFrame(
+                {name: getattr(npmodel, name) for name in metadata.array_props},
+                index=pdindex,
+            )
+
+        return pd.DataFrame(
+            {name: getattr(npmodel, name) for name in metadata.array_props}
+        )
+
+    @classmethod
+    def from_df(cls, df, model_cls: Type[NumpyDataModel]):
+        """This method converts a pandas dataframe to a numpy data model. The model must have only one dictionary-index"""
+        metadata: NumpyDataModelMetadata = model_cls._metadata  # type: ignore
+        assert len(metadata.index_props) == 1
+
+        index = cls.index_from_array(df.index.to_list())
+        kwargs = {name: df[name].to_numpy() for name in metadata.array_props}
+        kwargs[metadata.index_props[0]] = index
+        return model_cls(**kwargs)
+
+    @classmethod
+    def sync_index(cls, source: dict, target: dict):
+        """Sync two index dictionaries. The source index must be a subset of the target index.
+        Any entry in the target index that is not in the source index will be added to the source with empty range.
+        The index produced by this method does not have the range contiguous.
+        """
+        if len(target) == 0:
+            return
+
+        for key, value in target.items():
+            if isinstance(value, tuple):
+                if len(value) == 2:
+                    if key not in source:
+                        source[key] = (0, 0)
+                else:
+                    if key not in source:
+                        source[key] = (0, 0, {})
+                    cls.sync_index(source[key][2], value[2])
+            else:
+                if key not in source:
+                    source[key] = {}
+                cls.sync_index(source[key], value)
+
+    @classmethod
+    def offset_index(cls, index: dict, offset: int):
+        """This method offsets the index of a numpy data model by a given offset"""
+        for key, value in index.items():
+            if isinstance(value, tuple):
+                if len(value) == 2:
+                    index[key] = (value[0] + offset, value[1] + offset)
+                else:
+                    index[key] = (
+                        value[0] + offset,
+                        value[1] + offset,
+                        cls.offset_index(value[2], offset),
+                    )
+            else:
+                index[key] = cls.offset_index(value, offset)
+        return index
+
+    @classmethod
+    def is_range_index_valid(cls, npmodel: NumpyDataModel):
+        """Check if the index of a numpy data model is valid: the index is contiguous, no overlapping
+        and covers the entire range of the numpy arrays.
+
+        This function assumes that the index is a dictionary (or instance of Index[dict]), and each value in the index
+        is either (start, end) if the index is a single-level index or (start, end, nested_index) if the index is a
+        multi-level index.
+        """
+        metadata: NumpyDataModelMetadata = npmodel._metadata  # type: ignore
+        size = len(npmodel)
+        for name in metadata.index_props:
+            index = getattr(npmodel, name)
+            if isinstance(index, Index):
+                index = index.index
+            if not isinstance(index, dict):
+                raise TypeError(
+                    f"Index {name} is not a dictionary. This function only supports dictionary index"
+                )
+
+            # figure out the index level
+            index_level = cls._get_range_index_level(npmodel, index)
+            if index_level == 0 and size != 0:
+                raise ValueError("Index is empty but the numpy array is not empty")
+
+            indices = [index]
+            for level in tqdm(
+                range(index_level), desc="check index at different levels"
+            ):
+                cic = ContiguousIndexChecker()
+                next_indices = []
+                for idx in tqdm(indices):
+                    if len(idx) == 0:
+                        continue
+                    for elem in idx.values():
+                        cic.next(elem[0], elem[1])
+                        if len(elem) == 3:
+                            next_indices.append(elem[2])
+                if not cic.start == size:
+                    raise ValueError(
+                        f"Index is not covered all elements in the array at level {level}. Expected {size} but got {cic.start}"
+                    )
+                indices = next_indices
+
+    @classmethod
+    def _get_range_index_level(cls, npmodel: NumpyDataModel, index: dict):
+        """Get the level of a range index. This function assumes that the index is a dictionary, and each value in the index
+        is either (start, end) if the index is a single-level index or (start, end, nested_index) if the index is a
+        multi-level index."""
+        if len(index) == 0:
+            return 0
+
+        level = []
+        for key, value in index.items():
+            if not isinstance(value, (list, tuple)):
+                raise TypeError(f"Index element {value} is not a sequence")
+            if len(value) == 2:
+                if not isinstance(value[0], int) or not isinstance(value[1], int):
+                    raise TypeError(
+                        f"Index element {value} does not have the form: (start, end)"
+                    )
+                level.append(1)
+            elif len(value) == 3:
+                if (
+                    not isinstance(value[0], int)
+                    or not isinstance(value[1], int)
+                    and not isinstance(value[2], dict)
+                ):
+                    raise TypeError(
+                        f"Index element {value} does not have the form: (start, end, nested_index)"
+                    )
+                level.append(cls._get_range_index_level(npmodel, value[2]) + 1)
+            else:
+                raise TypeError(
+                    f"Index element {value} does not have the form: (start, end) or (start, end, nested_index)"
+                )
+
+        level = set(level)
+        if len(level) > 1:
+            raise ValueError(
+                "The index is not a balanced multi-level index as some elements have different levels"
+            )
+        return level.pop()
+
+    @classmethod
+    def to_nested_list(cls, arr: np.ndarray, index: dict):
+        """Convert a numpy array to a nested list. The nested level is determined by the index"""
+        out = []
+        for key, value in index.items():
+            if isinstance(value, tuple):
+                if len(value) == 2:
+                    out.append(arr[value[0] : value[1]])
+                else:
+                    out.append(cls.to_nested_list(arr, value[2]))
+            elif isinstance(value, dict):
+                out.append(cls.to_nested_list(arr, value))
+            else:
+                raise TypeError(
+                    f"Index element {value} does not have the form: (start, end) or (start, end, nested_index)"
+                )
+        return out
+
+    @classmethod
+    def index_to_array(cls, index: dict, prefix: tuple[str, ...] = tuple()):
+        """Convert a dictionary index to an array that is ready to be loaded into
+        pandas's multiindex via pd.MultiIndex.from_tuples
+        """
+        arr = []
+        for key, value in index.items():
+            if isinstance(value, (tuple, list)):
+                if len(value) == 2:
+                    for i in range(value[0], value[1]):
+                        arr.append(prefix + (key, i))
+                else:
+                    arr.extend(cls.index_to_array(value[2], prefix + (key,)))
+            elif isinstance(value, dict):
+                arr.extend(cls.index_to_array(value, prefix + (key,)))
+            else:
+                raise TypeError(
+                    f"Index element {value} does not have the form: (start, end) or (start, end, nested_index)"
+                )
+        return arr
+
+    @classmethod
+    def index_from_array(cls, arr: list[tuple]):
+        """Convert an array of tuples back to a dictionary index. The array is generated by index_to_array
+        or pandas.MultiIndex.to_list"""
+        index = {}
+        if len(arr) == 0:
+            return index
+
+        if not isinstance(arr[0], (tuple, list)) and len(arr[0]) == 1:
+            raise ValueError("Don't need a dictionary index for a single-level array")
+
+        pending_lst = []
+        for i, elem in enumerate(arr):
+            ptr = index
+            for j in range(len(elem) - 2):
+                if elem[j] not in ptr:
+                    ptr[elem[j]] = [i, i, {}]
+                    pending_lst.append((ptr, elem[j]))
+                item = ptr[elem[j]]
+                item[1] += 1
+                ptr = item[2]
+            if elem[-2] not in ptr:
+                ptr[elem[-2]] = [i, i]
+                pending_lst.append((ptr, elem[-2]))
+            ptr[elem[-2]][1] += 1
+
+        for obj, key in pending_lst:
+            obj[key] = tuple(obj[key])
+
+        return index
